/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "stdlib.h"
#include "atom_vec_granular.h"
#include "atom.h"
#include "domain.h"
#include "force.h"
#include "modify.h"
#include "fix.h"
#include "memory.h"
#include "error.h"

using namespace LAMMPS_NS;

#define DELTA 10000

/* ---------------------------------------------------------------------- */

AtomVecGranular::AtomVecGranular(LAMMPS *lmp, int narg, char **arg) :
  AtomVec(lmp, narg, arg)
{
  comm_x_only = comm_f_only = 0;
  size_comm = 9;
  size_reverse = 6;
  size_border = 14;
  size_data_atom = 7;
  size_data_vel = 7;
  xcol_data = 5;

  PI = 4.0*atan(1.0);
}

/* ----------------------------------------------------------------------
   grow atom arrays
   n = 0 grows arrays by DELTA
   n > 0 allocates arrays to size n 
------------------------------------------------------------------------- */

void AtomVecGranular::grow(int n)
{
  if (n == 0) nmax += DELTA;
  else nmax = n;
  atom->nmax = nmax;

  tag = atom->tag = (int *) 
    memory->srealloc(atom->tag,nmax*sizeof(int),"atom:tag");
  type = atom->type = (int *)
    memory->srealloc(atom->type,nmax*sizeof(int),"atom:type");
  mask = atom->mask = (int *) 
    memory->srealloc(atom->mask,nmax*sizeof(int),"atom:mask");
  image = atom->image = (int *) 
    memory->srealloc(atom->image,nmax*sizeof(int),"atom:image");
  x = atom->x = memory->grow_2d_double_array(atom->x,nmax,3,"atom:x");
  v = atom->v = memory->grow_2d_double_array(atom->v,nmax,3,"atom:v");
  f = atom->f = memory->grow_2d_double_array(atom->f,nmax,3,"atom:f");

  radius = atom->radius = (double *)
    memory->srealloc(atom->radius,nmax*sizeof(double),"atom:radius");
  density = atom->density = (double *)
    memory->srealloc(atom->density,nmax*sizeof(double),"atom:density");
  rmass = atom->rmass = (double *)
    memory->srealloc(atom->rmass,nmax*sizeof(double),"atom:rmass");

  phix = atom->phix = 
    memory->grow_2d_double_array(atom->phix,nmax,3,"atom:phix");
  phiv = atom->phiv = 
    memory->grow_2d_double_array(atom->phiv,nmax,3,"atom:phiv");
  phia = atom->phia =
    memory->grow_2d_double_array(atom->phia,nmax,3,"atom:phia");

  if (atom->nextra_grow)
    for (int iextra = 0; iextra < atom->nextra_grow; iextra++) 
      modify->fix[atom->extra_grow[iextra]]->grow_arrays(nmax);
}

/* ---------------------------------------------------------------------- */

void AtomVecGranular::reset_ptrs()
{
  tag = atom->tag;
  type = atom->type;
  mask = atom->mask;
  image = atom->image;
  x = atom->x;
  v = atom->v;
  f = atom->f;

  radius = atom->radius;
  density = atom->density;
  rmass = atom->rmass;
  phix = atom->phix;
  phiv = atom->phiv;
  phia = atom->phia;
}

/* ----------------------------------------------------------------------
   zero auxiliary data for owned atom I
   data in copy(), not including tag,type,mask,image,x,v
------------------------------------------------------------------------- */

void AtomVecGranular::zero_owned(int i)
{
  radius[i] = 0.0;
  density[i] = 0.0;
  rmass[i] = 0.0;
  phix[i][0] = 0.0;
  phix[i][1] = 0.0;
  phix[i][2] = 0.0;
  phiv[i][0] = 0.0;
  phiv[i][1] = 0.0;
  phiv[i][2] = 0.0;
}

/* ----------------------------------------------------------------------
   zero auxiliary data for n ghost atoms
   data in border(), not including x,tag,type,mask
   grow() is here since zero_ghost called first in hybrid::unpack_border()
------------------------------------------------------------------------- */

void AtomVecGranular::zero_ghost(int n, int first)
{
  int last = first + n;
  for (int i = first; i < last; i++) {
    if (i == nmax) atom->avec->grow(0);
    v[i][0] = 0.0;
    v[i][1] = 0.0;
    v[i][2] = 0.0;
    radius[i] = 0.0;
    rmass[i] = 0.0;
    phiv[i][0] = 0.0;
    phiv[i][1] = 0.0;
    phiv[i][2] = 0.0;
  }
}

/* ---------------------------------------------------------------------- */

void AtomVecGranular::copy(int i, int j)
{
  tag[j] = tag[i];
  type[j] = type[i];
  mask[j] = mask[i];
  image[j] = image[i];
  x[j][0] = x[i][0];
  x[j][1] = x[i][1];
  x[j][2] = x[i][2];
  v[j][0] = v[i][0];
  v[j][1] = v[i][1];
  v[j][2] = v[i][2];

  radius[j] = radius[i];
  density[j] = density[i];
  rmass[j] = rmass[i];
  phix[j][0] = phix[i][0];
  phix[j][1] = phix[i][1];
  phix[j][2] = phix[i][2];
  phiv[j][0] = phiv[i][0];
  phiv[j][1] = phiv[i][1];
  phiv[j][2] = phiv[i][2];

  if (atom->nextra_grow)
    for (int iextra = 0; iextra < atom->nextra_grow; iextra++) 
      modify->fix[atom->extra_grow[iextra]]->copy_arrays(i,j);
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::pack_comm(int n, int *list, double *buf, int *pbc_flags)
{
  int i,j,m;

  m = 0;
  if (pbc_flags[0] == 0) {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0];
      buf[m++] = x[j][1];
      buf[m++] = x[j][2];
      buf[m++] = v[j][0];
      buf[m++] = v[j][1];
      buf[m++] = v[j][2];
      buf[m++] = phiv[j][0];
      buf[m++] = phiv[j][1];
      buf[m++] = phiv[j][2];
    }
  } else {
    double xprd = domain->xprd;
    double yprd = domain->yprd;
    double zprd = domain->zprd;
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0] + pbc_flags[1]*xprd;
      buf[m++] = x[j][1] + pbc_flags[2]*yprd;
      buf[m++] = x[j][2] + pbc_flags[3]*zprd;
      buf[m++] = v[j][0];
      buf[m++] = v[j][1];
      buf[m++] = v[j][2];
      buf[m++] = phiv[j][0];
      buf[m++] = phiv[j][1];
      buf[m++] = phiv[j][2];
    }
  }
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::pack_comm_one(int i, double *buf)
{
  buf[0] = v[i][0];
  buf[1] = v[i][1];
  buf[2] = v[i][2];
  buf[3] = phiv[i][0];
  buf[4] = phiv[i][1];
  buf[5] = phiv[i][2];
  return 6;
}

/* ---------------------------------------------------------------------- */

void AtomVecGranular::unpack_comm(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    x[i][0] = buf[m++];
    x[i][1] = buf[m++];
    x[i][2] = buf[m++];
    v[i][0] = buf[m++];
    v[i][1] = buf[m++];
    v[i][2] = buf[m++];
    phiv[i][0] = buf[m++];
    phiv[i][1] = buf[m++];
    phiv[i][2] = buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::unpack_comm_one(int i, double *buf)
{
  v[i][0] = buf[0];
  v[i][1] = buf[1];
  v[i][2] = buf[2];
  phiv[i][0] = buf[3];
  phiv[i][1] = buf[4];
  phiv[i][2] = buf[5];
  return 6;
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::pack_reverse(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    buf[m++] = f[i][0];
    buf[m++] = f[i][1];
    buf[m++] = f[i][2];
    buf[m++] = phia[i][0];
    buf[m++] = phia[i][1];
    buf[m++] = phia[i][2];
  }
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::pack_reverse_one(int i, double *buf)
{
  buf[0] = phia[i][0];
  buf[1] = phia[i][1];
  buf[2] = phia[i][2];
  return 3;
}

/* ---------------------------------------------------------------------- */

void AtomVecGranular::unpack_reverse(int n, int *list, double *buf)
{
  int i,j,m;

  m = 0;
  for (i = 0; i < n; i++) {
    j = list[i];
    f[j][0] += buf[m++];
    f[j][1] += buf[m++];
    f[j][2] += buf[m++];
    phia[j][0] += buf[m++];
    phia[j][1] += buf[m++];
    phia[j][2] += buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::unpack_reverse_one(int i, double *buf)
{
  phia[i][0] = buf[0];
  phia[i][1] = buf[1];
  phia[i][2] = buf[2];
  return 3;
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::pack_border(int n, int *list, double *buf, int *pbc_flags)
{
  int i,j,m;

  m = 0;
  if (pbc_flags[0] == 0) {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0];
      buf[m++] = x[j][1];
      buf[m++] = x[j][2];
      buf[m++] = tag[j];
      buf[m++] = type[j];
      buf[m++] = mask[j];
      buf[m++] = v[j][0];
      buf[m++] = v[j][1];
      buf[m++] = v[j][2];
      buf[m++] = radius[j];
      buf[m++] = rmass[j];
      buf[m++] = phiv[j][0];
      buf[m++] = phiv[j][1];
      buf[m++] = phiv[j][2];
    }
  } else {
    double xprd = domain->xprd;
    double yprd = domain->yprd;
    double zprd = domain->zprd;
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0] + pbc_flags[1]*xprd;
      buf[m++] = x[j][1] + pbc_flags[2]*yprd;
      buf[m++] = x[j][2] + pbc_flags[3]*zprd;
      buf[m++] = tag[j];
      buf[m++] = type[j];
      buf[m++] = mask[j];
      buf[m++] = v[j][0];
      buf[m++] = v[j][1];
      buf[m++] = v[j][2];
      buf[m++] = radius[j];
      buf[m++] = rmass[j];
      buf[m++] = phiv[j][0];
      buf[m++] = phiv[j][1];
      buf[m++] = phiv[j][2];
    }
  }
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::pack_border_one(int i, double *buf)
{
  buf[0] = v[i][0];
  buf[1] = v[i][1];
  buf[2] = v[i][2];
  buf[3] = radius[i];
  buf[4] = rmass[i];
  buf[5] = phiv[i][0];
  buf[6] = phiv[i][1];
  buf[7] = phiv[i][2];
  return 8;
}

/* ---------------------------------------------------------------------- */

void AtomVecGranular::unpack_border(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    if (i == nmax) grow(0);
    x[i][0] = buf[m++];
    x[i][1] = buf[m++];
    x[i][2] = buf[m++];
    tag[i] = static_cast<int> (buf[m++]);
    type[i] = static_cast<int> (buf[m++]);
    mask[i] = static_cast<int> (buf[m++]);
    v[i][0] = buf[m++];
    v[i][1] = buf[m++];
    v[i][2] = buf[m++];
    radius[i] = buf[m++];
    rmass[i] = buf[m++];
    phiv[i][0] = buf[m++];
    phiv[i][1] = buf[m++];
    phiv[i][2] = buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::unpack_border_one(int i, double *buf)
{
  v[i][0] = buf[0];
  v[i][1] = buf[1];
  v[i][2] = buf[2];
  radius[i] = buf[3];
  rmass[i] = buf[4];
  phiv[i][0] = buf[5];
  phiv[i][1] = buf[6];
  phiv[i][2] = buf[7];
  return 8;
}

/* ----------------------------------------------------------------------
   pack data for atom I for sending to another proc
   xyz must be 1st 3 values, so comm::exchange() can test on them 
------------------------------------------------------------------------- */

int AtomVecGranular::pack_exchange(int i, double *buf)
{
  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];
  buf[m++] = tag[i];
  buf[m++] = type[i];
  buf[m++] = mask[i];
  buf[m++] = image[i];

  buf[m++] = radius[i];
  buf[m++] = density[i];
  buf[m++] = rmass[i];
  buf[m++] = phix[i][0];
  buf[m++] = phix[i][1];
  buf[m++] = phix[i][2];
  buf[m++] = phiv[i][0];
  buf[m++] = phiv[i][1];
  buf[m++] = phiv[i][2];

  if (atom->nextra_grow)
    for (int iextra = 0; iextra < atom->nextra_grow; iextra++) 
      m += modify->fix[atom->extra_grow[iextra]]->pack_exchange(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomVecGranular::unpack_exchange(double *buf)
{
  int nlocal = atom->nlocal;
  if (nlocal == nmax) grow(0);

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);

  radius[nlocal] = buf[m++];
  density[nlocal] = buf[m++];
  rmass[nlocal] = buf[m++];
  phix[nlocal][0] = buf[m++];
  phix[nlocal][1] = buf[m++];
  phix[nlocal][2] = buf[m++];
  phiv[nlocal][0] = buf[m++];
  phiv[nlocal][1] = buf[m++];
  phiv[nlocal][2] = buf[m++];

  if (atom->nextra_grow)
    for (int iextra = 0; iextra < atom->nextra_grow; iextra++) 
      m += modify->fix[atom->extra_grow[iextra]]->
	unpack_exchange(nlocal,&buf[m]);

  atom->nlocal++;
  return m;
}

/* ----------------------------------------------------------------------
   size of restart data for all atoms owned by this proc
   include extra data stored by fixes
------------------------------------------------------------------------- */

int AtomVecGranular::size_restart()
{
  int i;

  int nlocal = atom->nlocal;
  int n = 19 * nlocal;

  if (atom->nextra_restart)
    for (int iextra = 0; iextra < atom->nextra_restart; iextra++) 
      for (i = 0; i < nlocal; i++)
	n += modify->fix[atom->extra_restart[iextra]]->size_restart(i);

  return n;
}

/* ----------------------------------------------------------------------
   size of restart data for atom I
   do not include extra data stored by fixes, included by caller
------------------------------------------------------------------------- */

int AtomVecGranular::size_restart_one(int i)
{
  return 19;
}

/* ----------------------------------------------------------------------
   pack atom I's data for restart file including extra quantities
   xyz must be 1st 3 values, so that read_restart can test on them
   molecular types may be negative, but write as positive   
------------------------------------------------------------------------- */

int AtomVecGranular::pack_restart(int i, double *buf)
{
  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = tag[i];
  buf[m++] = type[i];
  buf[m++] = mask[i];
  buf[m++] = image[i];
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];

  buf[m++] = radius[i];
  buf[m++] = density[i];
  buf[m++] = phix[i][0];
  buf[m++] = phix[i][1];
  buf[m++] = phix[i][2];
  buf[m++] = phiv[i][0];
  buf[m++] = phiv[i][1];
  buf[m++] = phiv[i][2];

  if (atom->nextra_restart)
    for (int iextra = 0; iextra < atom->nextra_restart; iextra++) 
      m += modify->fix[atom->extra_restart[iextra]]->pack_restart(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ----------------------------------------------------------------------
   unpack data for one atom from restart file including extra quantities
------------------------------------------------------------------------- */

int AtomVecGranular::unpack_restart(double *buf)
{
  int nlocal = atom->nlocal;
  if (nlocal == nmax) {
    grow(0);
    if (atom->nextra_store)
      atom->extra = memory->grow_2d_double_array(atom->extra,nmax,
						 atom->nextra_store,
						 "atom:extra");
  }

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];

  radius[nlocal] = buf[m++];
  density[nlocal] = buf[m++];
  if (force->dimension == 3) 
    rmass[nlocal] = 4.0*PI/3.0 * 
      radius[nlocal]*radius[nlocal]*radius[nlocal] * density[nlocal];
  else
    rmass[nlocal] = PI * radius[nlocal]*radius[nlocal] * density[nlocal];

  phix[nlocal][0] = buf[m++];
  phix[nlocal][1] = buf[m++];
  phix[nlocal][2] = buf[m++];
  phiv[nlocal][0] = buf[m++];
  phiv[nlocal][1] = buf[m++];
  phiv[nlocal][2] = buf[m++];

  double **extra = atom->extra;
  if (atom->nextra_store) {
    int size = static_cast<int> (buf[0]) - m;
    for (int i = 0; i < size; i++) extra[nlocal][i] = buf[m++];
  }

  atom->nlocal++;
  return m;
}

/* ----------------------------------------------------------------------
   create one atom of type itype at x0,y0,z0
   set other values to defaults
------------------------------------------------------------------------- */

void AtomVecGranular::create_atom(int itype, double x0, double y0, double z0,
				  int ihybrid)
{
  int nlocal = atom->nlocal;
  if (nlocal == nmax) grow(0);

  tag[nlocal] = 0;
  type[nlocal] = itype;
  x[nlocal][0] = x0;
  x[nlocal][1] = y0;
  x[nlocal][2] = z0;
  mask[nlocal] = 1;
  image[nlocal] = (512 << 20) | (512 << 10) | 512;
  v[nlocal][0] = 0.0;
  v[nlocal][1] = 0.0;
  v[nlocal][2] = 0.0;

  radius[nlocal] = 0.5;
  density[nlocal] = 1.0;
  if (force->dimension == 3)
    rmass[nlocal] = 4.0*PI/3.0 *
      radius[nlocal]*radius[nlocal]*radius[nlocal] * density[nlocal];
  else
    rmass[nlocal] = PI * radius[nlocal]*radius[nlocal] * density[nlocal];
  phix[nlocal][0] = 0.0;
  phix[nlocal][1] = 0.0;
  phix[nlocal][2] = 0.0;
  phiv[nlocal][0] = 0.0;
  phiv[nlocal][1] = 0.0;
  phiv[nlocal][2] = 0.0;

  atom->nlocal++;
}

/* ----------------------------------------------------------------------
   unpack one line from Atoms section of data file
   initialize other atom quantities
------------------------------------------------------------------------- */

void AtomVecGranular::data_atom(double xtmp, double ytmp, double ztmp,
				int imagetmp, char **values, int ihybrid)
{
  int nlocal = atom->nlocal;
  if (nlocal == nmax) grow(0);

  tag[nlocal] = atoi(values[0]);
  if (tag[nlocal] <= 0)
    error->one("Invalid atom ID in Atoms section of data file");

  type[nlocal] = atoi(values[1]);
  if (type[nlocal] <= 0 || type[nlocal] > atom->ntypes)
    error->one("Invalid atom type in Atoms section of data file");

  radius[nlocal] = 0.5 * atof(values[2]);
  density[nlocal] = atof(values[3]);
  if (force->dimension == 3)
    rmass[nlocal] = 4.0*PI/3.0 *
      radius[nlocal]*radius[nlocal]*radius[nlocal] * density[nlocal];
  else
    rmass[nlocal] = PI * radius[nlocal]*radius[nlocal] * density[nlocal];

  x[nlocal][0] = xtmp;
  x[nlocal][1] = ytmp;
  x[nlocal][2] = ztmp;

  image[nlocal] = imagetmp;

  mask[nlocal] = 1;
  v[nlocal][0] = 0.0;
  v[nlocal][1] = 0.0;
  v[nlocal][2] = 0.0;
  phix[nlocal][0] = 0.0;
  phix[nlocal][1] = 0.0;
  phix[nlocal][2] = 0.0;
  phiv[nlocal][0] = 0.0;
  phiv[nlocal][1] = 0.0;
  phiv[nlocal][2] = 0.0;

  atom->nlocal++;
}

/* ----------------------------------------------------------------------
   unpack a single line from Velocity section of data file
------------------------------------------------------------------------- */

void AtomVecGranular::data_vel(int m, char *line, int ihybrid)
{
  int tmp;
  sscanf(line,"%d %lg %lg %lg %lg %lg %lg",
	 &tmp,&v[m][0],&v[m][1],&v[m][2],&phiv[m][0],&phiv[m][1],&phiv[m][2]);
}

/* ----------------------------------------------------------------------
   return # of bytes of allocated memory 
------------------------------------------------------------------------- */

int AtomVecGranular::memory_usage()
{
  int bytes = 0;

  if (atom->memcheck("tag")) bytes += nmax * sizeof(int);
  if (atom->memcheck("type")) bytes += nmax * sizeof(int);
  if (atom->memcheck("mask")) bytes += nmax * sizeof(int);
  if (atom->memcheck("image")) bytes += nmax * sizeof(int);
  if (atom->memcheck("x")) bytes += nmax*3 * sizeof(double);
  if (atom->memcheck("v")) bytes += nmax*3 * sizeof(double);
  if (atom->memcheck("f")) bytes += nmax*3 * sizeof(double);

  if (atom->memcheck("radius")) bytes += nmax * sizeof(double);
  if (atom->memcheck("density")) bytes += nmax * sizeof(double);
  if (atom->memcheck("rmass")) bytes += nmax * sizeof(double);
  if (atom->memcheck("phix")) bytes += nmax*3 * sizeof(double);
  if (atom->memcheck("phiv")) bytes += nmax*3 * sizeof(double);
  if (atom->memcheck("phia")) bytes += nmax*3 * sizeof(double);

  return bytes;
}
